package goodsutil;
/**
 * GoodsObjectFactory.java
 *
 *
 * Created: Tue Dec  4 12:06:41 2001
 *
 * @author <a href="mailto:hotari@iki.fi">Lari Hotari</a>
 * @version
 */

import java.io.*;
import java.util.*;
import java.util.Date;
import java.text.*;
import java.lang.reflect.*;
import goodslib.*;
import goodsjpi.*;
import oodblib.*;
import oodblib.adapter.*;

public abstract class GoodsObjectFactory implements OODBObjectFactory {
    private Thread shutdownthread;
    private Database db;
    private GoodsServer goodsserver;
    private long cachehits=0;
    private long cachemisses=0;

    protected GoodsObjectFactory () {
	initDatabase();
	shutdownthread=new Thread() {
		public void run() {
		    shutdown();
		}
	    };
	shutdownthread.setName("DB Shutdownhook");
	Runtime.getRuntime().addShutdownHook(shutdownthread);
    }

    // You'll have to implement these in the inherited class
    abstract public String getDatabaseName();
    abstract public File[] getDatabaseDirs();
    abstract public Properties getProperties();
    abstract protected Object getObjectFromCache(Object key);
    abstract protected void storeObjectInCache(Object key, Object value);
    abstract protected void removeObjectFromCache(Object key);

    public boolean isIntegratedServer() {
	String inproc=getProperties().getProperty("goodsdb.server.enabled");
	if(inproc!=null && (inproc.equalsIgnoreCase("false") || inproc.equalsIgnoreCase("off") || inproc.equals("0"))) {
	    return false;
	}
	return true;
    }

    public Database getDatabase() {
	return db;
    }

    public void deleteOODBObject(Object object) {

    }

    public void waitForCommit() {
	((Persistent)db.getRoot()).metaobject.waitForCommit();
    }

    public Transaction newTransaction() {
	return new GoodsTransactionAdapter((Persistent)db.getRoot());
    }

    public Transaction currentTransaction() {
	return new GoodsTransactionAdapter((Persistent)db.getRoot());
    }

    final public Object mapDataToNormal(Object obj) {
	if(obj==null) {
	    return null;
	}

	if(obj instanceof OODBObject) {
	    Object key=getObjectKey(obj);
	    if(key==null) {
		displayError("Cannot get objectid for " + obj);
		key=obj;
	    }
	    Object retval=getObjectFromCache(key);
	    if(retval==null) {
		if(obj instanceof PersistentHashMap) {
		    retval=new MapAdapter(this,(Map)obj);
		} else if (obj instanceof PersistentTreeSet) {
		    retval=new SortedSetAdapter(this,(SortedSet)obj);
		} else if (obj instanceof PersistentAbstractList) {
		    retval=new ListAdapter(this,(List)obj);
		} else if (obj instanceof PersistentStack) {
		    retval=new StackAdapter(this,(Stack)obj);
		} else {
		    retval=convertObjectToNormal(obj);
		}
		if(retval != null && retval != obj) {
		    cachemisses++;
		    //System.err.println("Cache missed! obj=" + obj);
		    storeObjectInCache(key,retval);
		} else {
		    retval=obj;
		}
	    } else {
		cachehits++;
	    }
	    return retval;
	} else {
	    return obj;
	}
    }

    final public Object mapNormalToData(Object obj) {
	if(obj==null) {
	    return null;
	}

	Object convertedobj;
	if(obj instanceof DataInOODB) {
	    convertedobj=((DataInOODB)obj).getOODBObject();
	} else if (obj instanceof MappedObject) {
	    convertedobj=((MappedObject)obj).getDataObject();
	} else {
	    convertedobj=convertObjectToData(obj);
	}
	return convertedobj;
    }

    public Object getObjectID(Object obj) {
	Object dataobj=mapNormalToData(obj);
	if(dataobj instanceof Persistent) {
	    waitForCommit();
	    int oid=((Persistent)dataobj).getOid();
	    if(oid != 0) {
		return new Integer(oid);
	    } else {
		displayError("The object id is 0 for dataobject " + dataobj);
		return null;
	    }
	} else {
	    return null;
	}
    }

    private void displayInfo(Object obj) {
	System.out.println(obj);
    }

    private void displayDebug(Object obj) {
	System.out.println(obj);	
    }

    private void displayError(Object obj) {

    }

    private void displayError(Object obj, Throwable t) {
	System.err.println(obj);
	if(t != null) 
	    t.printStackTrace(System.err);
    }


    protected Object convertObjectToNormal(Object obj) {
	return obj;
    }

    protected Object convertObjectToData(Object obj) {
	return obj;
    }

    public Object getObjectKey(Object obj) {
	Object oid=getObjectID(obj);
	if(oid==null) oid=new Integer(obj.hashCode());
	return oid;
    }

    public Object getNormalObjectByOid(Object oid) {
	Object cached=getObjectFromCache(oid);
	if(cached != null) {
	    return cached;
	} 
	Object dataobj=db.getObject(((Number)oid).intValue());
	if(dataobj != null) {
	    return mapDataToNormal(dataobj);
	} else {
	    return null;
	}
    }

    public Object getOODBObjectByOid(Object oid) {
	Object cached=getObjectFromCache(oid);
	if(cached != null) {
	    if(cached instanceof DataInOODB) {
		return ((DataInOODB)cached).getOODBObject();
	    }
	} 
	Object dataobj=db.getObject(((Number)oid).intValue());
	if(dataobj != null) {
	    return dataobj;
	} else {
	    return null;
	}
    }

    public void registerObject(Object obj) {
	storeObjectInCache(getObjectKey(obj),obj);
    }

    public void forgetObject(Object obj) {
	removeObjectFromCache(getObjectKey(obj));
    }

    private synchronized void initDatabase() {
	if(isIntegratedServer()) {
	    displayInfo("Starting integrated GOODS database... ");
	    GoodsLocalServer dbserver=new GoodsLocalServer(getDatabaseName(), getProperties(), getDatabaseDirs());
	    goodsserver=dbserver;
	    try {
		dbserver.start();
	    } catch (Exception e) {
		displayError("Could not start integrated GOODS database!",e);
	    }
	} else {
	    displayInfo("Connecting to remote GOODS database... ");
	    GoodsRemoteServer dbserver=new GoodsRemoteServer(getDatabaseName(), getProperties(), getDatabaseDirs());
	    goodsserver=dbserver;
	}
	db=new Database();
	if (!db.open((new File(getDatabaseDirs()[0], getDatabaseName() + ".cfg")).getAbsolutePath())) { 
            displayError("Could not open database!");
        } else {
	    displayInfo("OK");
	}
    }

    protected void shutdown() {
	closeDB();
    }

    public synchronized void closeDB() {
	if(db != null) {
	    displayInfo("Closing the db.");
	    try {
		db.close();
		db=null;
	    } catch (Exception e) {
		displayError("Error closing db.",e);
	    }
	}
	if(goodsserver != null) {
	    try {
		goodsserver.stop();
		goodsserver=null;
	    } catch (Exception e) {
		displayError("Error closing server.", e);
	    }
	}
    }

    public long getCacheHits() {
	return cachehits;
    }

    public long getCacheMisses() {
	return cachemisses;
    }

    public float getCacheRatio() {
	return cachehits / (cachehits+cachemisses+1);
    }

}// GoodsFactory
